//Uso questo Alarm perche' il timer di java (Swing)richiede la presenza di un interfaccia grafica
//e genera errori nell'esecuzione sui nodi remoti


//----------------------------------------------------------------------
// This code (c) 1998 Allen I. Holub. All rights reserved.
//----------------------------------------------------------------------
// This code may not be distributed by yourself except in binary form,
// incorporated into a java .class file. You may use this code freely
// for personal purposes, but you may not incorporate it into any
// commercial product without express permission of Allen I. Holub in writing.
//
// Sources code for this class can be found in the "Goodies" section of
// <http://www.holub.com>.
//----------------------------------------------------------------------


package MobileMonitoring;

import java.util.*;			// Need Date for testing
import java.awt.event.*;	// For ActionListener, etc.
import java.awt.*;			// For ActionListener, etc.

/**
 *	An interval timer. Once you "start" it, it runs for a predetermined
 *	amount of time. The timer runs on it's own, relatively high
 *  priority, thread while one or more other threads block, waiting for
 *  the timer to "expire."
 *  <p>	Three sorts of timers are supported:
 *	<dl>
 *	<dt> A "one-shot" timer
 *	<dd>	runs once, then expires. May be started again, but generally
 *			isn't.
 *	<dt> A "multi-shot" timer
 *	<dd>	works like a one shot, but expects to
 *			be started again, so is more efficient this situation.
 *			Must be stopped explicitly.
 *	<dt> A "continuous" (oscillator-style) timer
 *	<dd>	runs continuously. Starts up again automatically when
 *			it expires. Must be stopped explicitly.
 *  </dl>
 *	<p>All timers may be restarted (have the time interval set back to
 *	   the original value) while they are running.
 *
 * <p><b>Warnings:</b>
 * <ol>
 * <li>	It's easy to forget to stop() a multi-shot or continuous
 *		timer and create a memory leak as a consequence. 
 *		(The VM will keep the Thread object alive but
 *		suspended, even if there are no external references to it.)
 *		A finalizer is provided to throw an exception in this situation,
 *		but since the finalizer may not be called, that's not a perfect
 *		solution.
 * <li>	Calling wait() on a timer is possible, but it only works correctly
 *		for CONTINUOUS timers. The await() method works correctly
 *		for all types of timers, and is easier to use since you don't
 *		have to synchronize first. The only time that using a raw
 *		wait() makes sense if if you're interested in the
 *		InterruptedException, which is silently absorbed by await().
 * </ol>
 **/

public class Alarm
{ 										// A compiler bug (1.1x, 1.2rc1)
	private int  			delay;		// permit blank finals.
	private Alarm.Clock 	clock		= null;
	private boolean 		is_stopped	= false;

	public static class Mode { private Mode(){} }

	/****************************************************************
	 * CONTINUOUS alarms run continuosly (until stopped), sending
	 * out notification messages and releasing waiting threads
	 * at a regular interval.
	 */
	public static final Mode CONTINUOUS  = new Mode();

	/****************************************************************
	 *	ONE_SHOT alarms run once, then stop automatically. [You do not
	 *	have to call stop().] Can be started again manually by calling
	 *  start(), but a new internal thread 	must be created to do so.
	 */
	public static final Mode ONE_SHOT    = new Mode();

	/****************************************************************
	 * MULTI_SHOT alarms work just like ONE_SHOT alarms, but they
	 * can be restarted more efficiently. (The timer thread for a
	 * MULTI_SHOT alarm is not destroyed until you stop() the alarm.)
	 * Use a MULTI_SHOT when you know that the one-shot will
	 * be restarted at regular intervals and don't mind having an extra
	 * thread kicking around in a suspened state.
	 */
	public static final Mode MULTI_SHOT  = new Mode();
	
	private Mode type;

	//------------------------------------------------------------------

    private ActionListener	observers	= null;
	private Thread 			notifier;

	/****************************************************************
	 * "Action command" sent to the ActionListeners when the timer
	 * has been stopped manually.
	 */
    public static final String STOPPED = "stopped";

	/****************************************************************
	 * "Action command" sent to the ActionListeners when the timer
	 * has expired or "ticked."
	 */
    public static final String EXPIRED = "expired";


	/****************************************************************
	 * Create a timer that expires after the indicated delay
	 *  in milliseconds.
	 *
	 *  <p> 
	 *  Generally, timers must run at high priority to prevent
	 *  the waiting threads from starting up immediately on being
	 *  notified, thereby messing up the timing. If a high-priority
	 *  thread is waiting on a recurring timer, however, the next
	 *	"tick" could be delayed by the amount of time the high-priority
	 *	thread spends doing whatever it does after being notified.
	 *  Alarms run at the highest priority permitted by the thread
	 *  group of which they are a member, but spend most of their
	 *  time sleeping.
	 *
	 * @param <b>delay</b>	Time to expiration (nominal) in milliseconds.
	 * @param <b>type</b>	One of the following symbolic constants (see):
	 *	<table>
	 *	<tr><td valign=top>{@link CONTINUOUS}</td><td valign=top>
	 *									The timer runs continuously
	 *									Must call stop() when you're
	 *									done with it.</td></tr>
	 *	<tr><td valign=top>{@link ONE_SHOT}</td>	<td valign=top>
	 *									The timer runs once, then stops
	 *									automatically. You do not
	 *									have to call stop().</td></tr>
	 *	<tr><td valign=top>{@link MULTI_SHOT}</td><td valign=top>
	 *									Like ONE_SHOT, but can be
	 *								    restarted more efficiently.</td></tr>
	 *	</table>
	 *	If this argument is <code>null</code>, <code>CONTINUOUS</code>
	 *  is used.
	 **/

	public Alarm( int delay, Mode type )
	{	this.delay 	= delay;
		this.type	= (type == null) ? CONTINUOUS : type ;
	}

	/****************************************************************
	 * Make a continuous timer.
	 */

	public Alarm( int delay )
	{	this( delay, CONTINUOUS );
	}

	/****************************************************************
	 *  Start up a timer or restart an expired timer. If the timer is
	 *  running, it is set back to the original count without releasing
	 *  any of the threads that are waiting for it to expire.
	 *  (For example, if you start up a 10-second timer and then restart
	 *  it after 5 seconds, the waiting threads won't be notified until
	 *  10 seconds after the restart---15 seconds after the original
	 *  start time.)
	 *  <p>
	 *	Starting a running timer causes a new thread to be created.
	 **/

	public synchronized void start()
	{	if( clock != null )
		{	if( type == MULTI_SHOT && clock.has_expired() )
			{
				clock.restart();
				return;
			}
			else
			{	clock.no_notifications();
	        	clock.interrupt();
			}
		}

        clock = new Clock();
		clock.start();
		is_stopped = false;
	}

	/****************************************************************
	 *	Stops the current timer abruptly, releasing all waiting
	 *  threads. The timer can be restarted by calling start().
	 *  There's no good way for a thread to determine if it was notified
	 *  as the result of a stop() or a normal expiration.
	 **/

	public synchronized void stop()
	{	if( clock != null )	
			clock.interrupt();

		clock 	   = null;
		is_stopped = true;
		notifyAll();
	}

	/****************************************************************
	 * It is a bug not to stop() a CONTINUOUS or MULTI_SHOT timer 
	 * when you're done with it. This finalizer helps you detect the
	 * bug by throwing an Error() if the timer that's being destroyed
	 * is still running.
	 **/

	public void finalize()
	{	if( clock != null )
			throw new Error("Alarm was not stopped before being destroyed");
	}

	/****************************************************************
	 * A long time (roughly 292,271,023 years) that you can use for
	 * the timeout value in {@link await()}.`
	 **/

	public static final long FOREVER = Long.MAX_VALUE;

	/****************************************************************
	 *	Wait for the timer to expire. Returns immediately in the case
	 *	of an expired ONE_SHOT or MULTI_SHOT timer.  Blocks until the
	 *	timer expires in all other situations (including any sort of
	 *	timer that has not yet been started).
	 *
	 *  @return false if the method returned because the timer was
	 *			stopped, true if the timer simply expired.
	 *
	 *	@see FOREVER
	 **/

	public synchronized boolean await( long timeout )
	{
		if( clock == null || !clock.has_expired() )
		{	try								{ wait( timeout );	}
			catch( InterruptedException e ) { /*do nothing*/	}
		}
		return !is_stopped;
	}

	/****************************************************************
	 * Same as <code>await(Alarm.FOREVER)</code>
	 **/
	public boolean await(){ return await( FOREVER ); }


	/****************************************************************
	 *	Add a listener that will be notified the next time that the
	 *	Alarm goes off. The listeners are notified on
	 *	a thread that's created just for that purpose, rather than
	 *	being notified from the timer thread. This way the time
	 *	spent doing notification dosn't impact the time interval
	 *	used by the timer. The "action command" in the ActionEvent
	 *	object will be either the String "stopped" or "expired"
	 *	(which are also defined in the symbolic constants {@link Alarm.STOPPED}
	 *	and {@link Alarm.EXPIRED}), depending on whether this notification
	 *  occured because the timer was stopped manually, or because
	 *	it expired in the normal way.
	 */

    public synchronized void addActionListener(ActionListener l)
    {	observers = AWTEventMulticaster.add(observers, l);
		if( notifier == null )
		{	notifier = new Alarm.Notifier();
			notifier.setDaemon( true );
			notifier.start();
		}
    }

	/****************************************************************
	 * Remove a listener
	 */
		
    public void removeActionListener(ActionListener l)
    {	observers = AWTEventMulticaster.remove(observers, l);
    }


	public class Notifier extends Thread
	{
		public void run()
		{
			ActionListener copy;

			while( observers != null )							//#notifier_loop
			{	synchronized( Alarm.this )
				{	try
					{	Alarm.this.wait();						//#notifier_wait
					}
					catch(InterruptedException e){/*ignore*/}

					if( (copy = observers) == null)				//#notifier_copy
						notifier = null;						//#notifier_clear
	 		}

				if( copy != null )								//#notifier_check
					copy.actionPerformed( 						//#notifier_action
							new ActionEvent(this,0,
									is_stopped ? STOPPED : EXPIRED));
			}

			// can't set notifier=null here
		}
	}

	//================================================================
	// Support classes:
	//================================================================
	
	private final class Clock extends Thread
    {
		private boolean expired 			= false;	// continuous timers don't expire
		private boolean notifications_off	= false;

		Clock()
		{	setPriority( getThreadGroup().getMaxPriority() );
		}


		public void run()
		{	
			while( !isInterrupted() )
			{	try
				{	sleep(delay); // release the monitor while sleeping

					if( isInterrupted() )	// don't notify waiting threads if
						break;				// we've been stopped by the
											// Alarm object.
					synchronized( this ) 					//#clock_synch
					{	
						expired = true;
						if( !notifications_off )
						{	synchronized( Alarm.this )
							{	Alarm.this.notifyAll();		//#clock_notify
							}
						}

						if( type == MULTI_SHOT )
		                {	wait();						// suspend		//#clock_suspend
						}
						else if( type == ONE_SHOT )
	                    {
							// Set the outer-class reference to the
							// current clock to null so that the
							// memory can be reclaimed. Note that
							// we're still synchronized on the outer-
							// class object.
																//#above_destroy
							synchronized( Alarm.this)
							{	if( Alarm.this.clock == this )
									Alarm.this.clock = null;	//#clock_destroy
							}
							break;
						}
					}
				}
				catch(InterruptedException e) // don't notify the waiting
				{	break;					  // threads because an
				}							  // interrupt is used to stop
			}								  // the timer.
		}

		public void no_notifications()
		{	notifications_off = true;
		}

		public synchronized void restart()
		{	expired = false;
			notify();						// resume			//#clock_release
		}

		public boolean has_expired()				// CONTINUOUS
		{	return (type != CONTINUOUS) && expired; // timers never
		}											// expire.
	};

	
}
